from distutils.core import setup 

setup(
    name='ach',
    author='Travis Hathaway',
    author_email='travis.j.hathaway@gmail.com',
    version='0.1',
    packages=[
        'ach',
    ],
    url='https://github.com/travishathaway/python-ach',
    license='MIT License',
    long_description=open('README.md').read(),
)
